{
=============================================================================
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  |
| AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    |
| IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   |
| ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR  |
| ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       |
| DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR   |
| SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER   |
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT           |
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY    |
| OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  |
| DAMAGE.                                                                      |
=============================================================================
   Author:  D.J.C. Blijker.
   Date:    13-4-2003
   version: 1.0
   e-mail:  c.blijker@wxs.nl
=============================================================================
  This unit will work in Delphi as well in Kylix!!!
=============================================================================

 Main target of this unit is to supply applications with an easy-to-use and
 consistent means of finding the exact network-address, port and protocol
 of (self implemented) network services.

 This protocol works in tight cooperation with the ServiceSeeker and
 the THxServiceProvider class.

==============================================================================
History:

  version: 1.0  Initial release.
==============================================================================
}
unit ServiceBroker_logicLayer;

interface

uses
{$IFDEF LINUX}
  QExtCtrls,
{$ELSE}
  extctrls,
{$ENDIF}
  SysUtils, HxTcpIpcUdp, HxDataExchange, ServiceBroker_Shared;
type
  THxServiceBroker = class
  private
    udp: THxUdp;
    de: THxDataExchange;
    lstServices: THxServiceLongList;
    tmr: TTimer;
    procedure HandleServiceRequest( ptr: pchar; iSize: integer );
    procedure HandleNewService( ptr: pchar; iSize: integer );
    procedure SendAnswer( sServiceName, sAddress, sPort: string;
                          eProtocol: THxProtocol; iMessNr: integer );
    procedure SendNewServiceAck( sServiceName, sAddress, sPort: string;
                                 eProtocol: THxProtocol );
    procedure HandleDataAvailable( ptr: pchar; iSize: integer;
                                   sRemoteIP, sPort: string );
    procedure SendRequestAnswer( sServiceName, sAddress, sPort: string;
                                 eProtocol: THxProtocol );
    procedure AddService( sNameOfService, sPort: string;
                          eProtocol: THxProtocol ); overload;
    procedure RemoveService( sNameOfService, sAddressOfService, sPort: string;
                             eProtocol: THxProtocol );
    procedure HandleTimer( Sender: TObject);
    procedure HandleRevokeService( ptr: pchar; iSize: integer );
    function  GetNrOfServices: integer;
    function  GetServiceInfo( index: integer ): PHxServiceLong;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddService( sNameOfService, sAddressOfService, sPort: string;
                          eProtocol: THxProtocol ); overload;
    property NrOfServices: integer read GetNrOfServices;
    property ServiceInfo[index: integer]: PHxServiceLong read GetServiceInfo;
  end;


implementation



{ THxServiceBroker }

constructor THxServiceBroker.Create;
begin
  udp := THxUdp.Create;
  udp.ListeningPort := SERVICEBROKER_PORT;
  udp.OnDataAvailable := HandleDataAvailable;

  de := THxDataExchange.Create;

  lstServices := THxServiceLongList.Create;

// All service providers must confirm them self at regular intervals. If a
// service provider fails to do so the service will be removed from the list.
  tmr := TTimer.Create( nil );
  tmr.Interval := 2000;
  tmr.OnTimer := HandleTimer;
  tmr.Enabled := True;

// The service broker acts as a service provider as well..
  AddService( SERVICEBROKER_SERVICE, SERVICEBROKER_PORT, pcUdp );
end;

destructor THxServiceBroker.Destroy;
begin
  udp.Free;
  de.Free;
  lstServices.Free;

  inherited;
end;

procedure THxServiceBroker.HandleTimer( Sender: TObject );
var
  i: integer;
begin
  i := lstServices.Count-1;
  while i > 0 do  // Leave the first element because that's our own service..
  begin
    Dec( lstServices[i]^.iTimeToLive, 2 );
    if lstServices[i]^.iTimeToLive <= 0 then
      lstServices.Delete( i );
    Dec( i );
  end;
end;

procedure THxServiceBroker.AddService( sNameOfService, sPort: string;
                                       eProtocol: THxProtocol );
var
  sLocalAddr: string;
begin
  sLocalAddr := GetLocalAddress;

  AddService( sNameOfService, sLocalAddr, sPort, eProtocol );
end;

procedure THxServiceBroker.AddService( sNameOfService,
                                       sAddressOfService,
                                       sPort: string;
                                       eProtocol: THxProtocol );
var
  pService: PHxServiceLong;
begin
  if( sNameOfService = '' )or( sPort = '' )then
    raise Exception.Create('Attempt to add an invalid service.');

// Make sure only one version is in the list..
  RemoveService( sNameOfService, sAddressOfService, sPort, eProtocol );

  New( pService );
  lstServices.Add( pService );

  pService^.iTimeToLive := 60; // one minute..
  pService^.sServiceName := sNameOfService;
  pService^.sAddress := sAddressOfService;
  pService^.sPort := sPort;
  pService^.eProtocolType := eProtocol;
end;

procedure THxServiceBroker.HandleDataAvailable( ptr: pchar; iSize: integer;
                                                sRemoteIP, sPort: string );
var
  sSAPHeader: string;
  iMessNr: integer;
begin
  try
    ptr := de.ReadString( ptr, sSAPHeader, iSize );
    if sSAPHeader <> SAP_HEADER then
      exit; // ignore 'strange' messages..
    ptr := de.ReadInteger( ptr, iMessNr, iSize );
    case iMessNr of
      INVALID_NUMBER: exit; // ignore invalid messages..
      NEW_SERVICE_ADDRESS: HandleNewService( ptr, iSize );
      SERVICE_REQ: HandleServiceRequest( ptr, iSize );
      REVOKE_SERVICE: HandleRevokeService( ptr, iSize );
    else
      exit; // ignore not supported messages..
    end;
  except
    // ignore also..
  end;
end;

procedure THxServiceBroker.HandleNewService( ptr: pchar; iSize: integer );
var
  sServAddr, sServPort, sServName: string;
  eProtocol: THxProtocol;
  iProtocol: integer;
begin
  if iSize > 0 then
    ptr := de.ReadString( ptr, sServName, iSize );

  if iSize > 0 then
    ptr := de.ReadString( ptr, sServAddr, iSize );

  if iSize > 0 then
    ptr := de.ReadString( ptr, sServPort, iSize );

  iProtocol := -1;  // invalid protocol...
  if iSize > 0 then
    de.ReadInteger( ptr, iProtocol, iSize );

  if( iSize = 0 )and
  ( iProtocol >= 0 )and
  ( iProtocol <= integer( High( THxProtocol )))then
  begin
    eProtocol := THxProtocol( iProtocol );
    AddService( sServName, sServAddr, sServPort, eProtocol );
    SendNewServiceAck( sServName, sServAddr, sServPort, eProtocol );
  end;
end;

procedure THxServiceBroker.HandleRevokeService( ptr: pchar; iSize: integer );
var
  sServAddr, sServPort, sServName: string;
  eProtocol: THxProtocol;
  iProtocol: integer;
begin
  if iSize > 0 then
    ptr := de.ReadString( ptr, sServName, iSize );

  if iSize > 0 then
    ptr := de.ReadString( ptr, sServAddr, iSize );

  if iSize > 0 then
    ptr := de.ReadString( ptr, sServPort, iSize );

  iProtocol := -1;  // invalid protocol...
  if iSize > 0 then
    de.ReadInteger( ptr, iProtocol, iSize );

  if( iSize = 0 )and
  ( iProtocol >= 0 )and
  ( iProtocol <= integer( High( THxProtocol )))then
  begin
    eProtocol := THxProtocol( iProtocol );
    RemoveService( sServName, sServAddr, sServPort, eProtocol );
  end;
end;


procedure THxServiceBroker.HandleServiceRequest( ptr: pchar; iSize: integer );
var
  i, iProtocol: integer;
  sReqService: string;
  eProtocol: THxProtocol;
begin
  if iSize > 0 then
    ptr := de.ReadString( ptr, sReqService, iSize );

  iProtocol := -1;
  if iSize > 0 then
    de.ReadInteger( ptr, iProtocol, iSize );

  if( iSize = 0 )and
  ( iProtocol >= 0 )and
  ( iProtocol <= integer( High( THxProtocol )))then
  begin
    eProtocol := THxProtocol( iProtocol );

    for i := 0 to lstServices.Count -1 do
      if( lstServices[i]^.sServiceName = sReqService )and
      ( eProtocol = lstServices[i]^.eProtocolType )then
        SendRequestAnswer( sReqService, lstServices[i]^.sAddress,
                           lstServices[i]^.sPort, eProtocol );
  end;
end;

procedure THxServiceBroker.RemoveService( sNameOfService,
                                          sAddressOfService,
                                          sPort: string;
                                          eProtocol: THxProtocol );
var
  i: integer;
begin
  for i := 0 to lstServices.Count -1 do
    if( lstServices[i]^.sServiceName = sNameOfService )and
    ( lstServices[i]^.sPort = sPort )and
    ( lstServices[i]^.eProtocolType = eProtocol )and
    ( lstServices[i]^.sAddress = sAddressOfService )then
    begin
      lstServices.Delete( i );
      break;
    end;
end;

procedure THxServiceBroker.SendRequestAnswer( sServiceName, sAddress,
                                              sPort: string;
                                              eProtocol: THxProtocol );
begin
  SendAnswer( sServiceName, sAddress, sPort, eProtocol, SERVICE_ANS );
end;

procedure THxServiceBroker.SendNewServiceAck( sServiceName, sAddress,
                                              sPort: string;
                                              eProtocol: THxProtocol );
begin
  SendAnswer( sServiceName, sAddress, sPort, eProtocol, ACK_SERVICE_ADDRESS );
end;

procedure THxServiceBroker.SendAnswer( sServiceName, sAddress, sPort: string;
                                       eProtocol: THxProtocol;
                                       iMessNr: integer );
var
  ptr: pchar;
  iSize: integer;
begin
  de.StartCreatingMessage;
  de.AddString( SAP_HEADER );
  de.AddInteger( iMessNr );
  de.AddString( sServiceName );
  de.AddString( sAddress );
  de.AddString( sPort );
  de.AddInteger( integer( eProtocol ));
  de.GetCreatedMessage( ptr, iSize );

  udp.Reply( ptr, iSize );
end;

function THxServiceBroker.GetNrOfServices: integer;
begin
  result := lstServices.Count;
end;

function THxServiceBroker.GetServiceInfo( index: integer ): PHxServiceLong;
begin
  if( index >= 0 )and( index < lstServices.Count )then
    result := lstServices[index]
  else
    result := nil;
end;

end.
